part of sliders;

class _RenderBaseSlider extends RenderProxyBox
    with RelayoutWhenSystemFontsChangeMixin {
  _RenderBaseSlider({
    dynamic min,
    dynamic max,
    double interval,
    double stepSize,
    SliderStepDuration stepDuration,
    int minorTicksPerInterval,
    bool showTicks,
    bool showLabels,
    bool showDivisors,
    bool showTooltip,
    LabelPlacement labelPlacement,
    NumberFormat numberFormat,
    DateFormat dateFormat,
    DateIntervalType dateIntervalType,
    LabelFormatterCallback labelFormatterCallback,
    TooltipTextFormatterCallback tooltipTextFormatterCallback,
    SfTrackShape trackShape,
    SfDivisorShape divisorShape,
    SfOverlayShape overlayShape,
    SfThumbShape thumbShape,
    SfTickShape tickShape,
    SfTickShape minorTickShape,
    SfTooltipShape tooltipShape,
    SfSliderThemeData sliderThemeData,
    TextDirection textDirection,
    MediaQueryData mediaQueryData,
  })  : _min = min,
        _max = max,
        _interval = interval,
        _stepSize = stepSize,
        _stepDuration = stepDuration,
        _minorTicksPerInterval = minorTicksPerInterval,
        _showTicks = showTicks,
        _showLabels = showLabels,
        _showDivisors = showDivisors,
        _showTooltip = showTooltip,
        _labelPlacement = labelPlacement,
        _numberFormat = numberFormat,
        _dateFormat = dateFormat,
        _dateIntervalType = dateIntervalType,
        _labelFormatterCallback = labelFormatterCallback,
        _tooltipTextFormatterCallback = tooltipTextFormatterCallback,
        _trackShape = trackShape,
        _divisorShape = divisorShape,
        _overlayShape = overlayShape,
        _thumbShape = thumbShape,
        _tickShape = tickShape,
        _minorTickShape = minorTickShape,
        _tooltipShape = tooltipShape,
        _sliderThemeData = sliderThemeData,
        _textDirection = textDirection,
        _mediaQueryData = mediaQueryData {
    _maxTrackHeight = _getMaxTrackHeight();
    _trackOffset = _getTrackOffset();

    if (isDateTime) {
      _minInms = min.millisecondsSinceEpoch.toDouble();
      _maxInms = max.millisecondsSinceEpoch.toDouble();
    }

    _visibleLabels = <String>[];
    _majorTickPositions = <double>[];
    _minorTickPositions = <double>[];
  }

  double _minInms;

  double _maxInms;

  double _divisions;

  //ignore: prefer_final_fields
  bool _willDrawTooltip = false;

  //ignore: prefer_final_fields
  bool _isInteractionEnd = true;

  List<String> _visibleLabels;

  List<double> _majorTickPositions;

  List<double> _minorTickPositions;

  List<double> _unformattedLabels;

  HorizontalDragGestureRecognizer _dragGestureRecognizer;

  TapGestureRecognizer _tapGestureRecognizer;

  final TextPainter _textPainter = TextPainter();

  double _actualHeight;

  Offset _trackOffset;

  double _maxTrackHeight;

  // It stores the current touch x-position, which is used in
  // the [endInteraction] and [dragUpdate] method.
  //ignore: prefer_final_fields
  double _currentX = 0.0;

  bool get isDateTime =>
      _min.runtimeType == DateTime && _max.runtimeType == DateTime;

  double get actualMin => isDateTime ? _minInms : _min;

  double get actualMax => isDateTime ? _maxInms : _max;

  SfThumb _activeThumb;

  bool get isDiscrete =>
      (stepSize != null && stepSize > 0) || (stepDuration != null);

  Size get actualDivisorSize =>
      _divisorShape.getPreferredSize(_sliderThemeData);

  Size get actualTickSize => _tickShape.getPreferredSize(_sliderThemeData);

  Size get actualMinorTickSize =>
      _minorTickShape.getPreferredSize(_sliderThemeData);

  Size get actualLabelSize => Size.fromHeight(math.max(
      _sliderThemeData.inactiveLabelStyle.fontSize,
      _sliderThemeData.activeLabelStyle.fontSize));

  Rect get trackRect =>
      _trackShape.getPreferredRect(this, _sliderThemeData, Offset.zero);

  Size get actualThumbSize => _thumbShape.getPreferredSize(_sliderThemeData);

  Size get actualOverlaySize =>
      _overlayShape.getPreferredSize(_sliderThemeData);

  double get _actualTickHeight => _showTicks
      ? _sliderThemeData.tickSize.height +
          (_sliderThemeData.tickOffset != null
              ? _sliderThemeData.tickOffset.dy
              : 0)
      : 0;

  double get _actualMinorTickHeight =>
      _minorTicksPerInterval != null ? actualMinorTickSize.height : 0;

  double get _actualLabelHeight => _showLabels
      ? actualLabelSize.height * _textPainter.textScaleFactor +
          (_sliderThemeData.labelOffset != null
              ? _sliderThemeData.labelOffset.dy
              : 0)
      : 0;

  final double _minTrackWidth = kMinInteractiveDimension * 3;

  // Here 10 is a gap between tooltip nose and thumb.
  double get tooltipStartY => _tooltipShape is SfPaddleTooltipShape
      ? math.max(actualThumbSize.height, trackRect.height) / 2
      : math.max(actualThumbSize.height, trackRect.height) / 2 + 10;

  dynamic get min => _min;
  dynamic _min;
  set min(dynamic value) {
    if (_min == value) {
      return;
    }
    _min = value;

    if (_min is DateTime) {
      _minInms = _min.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  dynamic get max => _max;
  dynamic _max;
  set max(dynamic value) {
    if (_max == value) {
      return;
    }
    _max = value;

    if (_max is DateTime) {
      _maxInms = _max.millisecondsSinceEpoch.toDouble();
    }
    markNeedsPaint();
  }

  double get interval => _interval;
  double _interval;

  set interval(double value) {
    if (_interval == value) {
      return;
    }
    _interval = value;
    markNeedsPaint();
  }

  double get stepSize => _stepSize;
  double _stepSize;

  set stepSize(double value) {
    if (_stepSize == value) {
      return;
    }
    _stepSize = value;
    markNeedsPaint();
  }

  SliderStepDuration get stepDuration => _stepDuration;
  SliderStepDuration _stepDuration;
  set stepDuration(SliderStepDuration value) {
    if (_stepDuration == value) {
      return;
    }
    _stepDuration = value;
  }

  int get minorTicksPerInterval => _minorTicksPerInterval;
  int _minorTicksPerInterval;

  set minorTicksPerInterval(int value) {
    if (_minorTicksPerInterval == value) {
      return;
    }
    _minorTicksPerInterval = value;
    markNeedsPaint();
  }

  bool get showTicks => _showTicks;
  bool _showTicks;

  set showTicks(bool value) {
    if (_showTicks == value) {
      return;
    }
    _showTicks = value;
    markNeedsLayout();
  }

  bool get showLabels => _showLabels;
  bool _showLabels;

  set showLabels(bool value) {
    if (_showLabels == value) {
      return;
    }
    _showLabels = value;
    markNeedsLayout();
  }

  bool get showDivisors => _showDivisors;
  bool _showDivisors;

  set showDivisors(bool value) {
    if (_showDivisors == value) {
      return;
    }
    _showDivisors = value;
    markNeedsLayout();
  }

  bool get showTooltip => _showTooltip;
  bool _showTooltip;
  set showTooltip(bool value) {
    if (_showTooltip == value) {
      return;
    }
    _showTooltip = value;
  }

  LabelPlacement get labelPlacement => _labelPlacement;
  LabelPlacement _labelPlacement;

  set labelPlacement(LabelPlacement value) {
    if (_labelPlacement == value) {
      return;
    }
    _labelPlacement = value;
    markNeedsPaint();
  }

  NumberFormat get numberFormat => _numberFormat;
  NumberFormat _numberFormat;

  set numberFormat(NumberFormat value) {
    if (_numberFormat == value) {
      return;
    }
    _numberFormat = value;
    markNeedsPaint();
  }

  DateIntervalType get dateIntervalType => _dateIntervalType;
  DateIntervalType _dateIntervalType;

  set dateIntervalType(DateIntervalType value) {
    if (_dateIntervalType == value) {
      return;
    }
    _dateIntervalType = value;
    markNeedsPaint();
  }

  DateFormat get dateFormat => _dateFormat;
  DateFormat _dateFormat;

  set dateFormat(DateFormat value) {
    if (_dateFormat == value) {
      return;
    }
    _dateFormat = value;
    markNeedsPaint();
  }

  LabelFormatterCallback get labelFormatterCallback => _labelFormatterCallback;
  LabelFormatterCallback _labelFormatterCallback;

  set labelFormatterCallback(LabelFormatterCallback value) {
    if (_labelFormatterCallback == value) {
      return;
    }
    _labelFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  TooltipTextFormatterCallback get tooltipTextFormatterCallback =>
      _tooltipTextFormatterCallback;
  TooltipTextFormatterCallback _tooltipTextFormatterCallback;

  set tooltipTextFormatterCallback(TooltipTextFormatterCallback value) {
    if (_tooltipTextFormatterCallback == value) {
      return;
    }
    _tooltipTextFormatterCallback = value;
    markNeedsSemanticsUpdate();
  }

  SfThumbShape get thumbShape => _thumbShape;
  SfThumbShape _thumbShape;

  set thumbShape(SfThumbShape value) {
    if (_thumbShape == value) {
      return;
    }
    _thumbShape = value;
    markNeedsLayout();
  }

  SfOverlayShape get overlayShape => _overlayShape;
  SfOverlayShape _overlayShape;

  set overlayShape(SfOverlayShape value) {
    if (_overlayShape == value) {
      return;
    }
    _overlayShape = value;
    markNeedsLayout();
  }

  SfTrackShape get trackShape => _trackShape;
  SfTrackShape _trackShape;

  set trackShape(SfTrackShape value) {
    if (_trackShape == value) {
      return;
    }
    _trackShape = value;
    markNeedsLayout();
  }

  SfDivisorShape get divisorShape => _divisorShape;
  SfDivisorShape _divisorShape;

  set divisorShape(SfDivisorShape value) {
    if (_divisorShape == value) {
      return;
    }
    _divisorShape = value;
    markNeedsLayout();
  }

  SfTickShape get tickShape => _tickShape;
  SfTickShape _tickShape;

  set tickShape(SfTickShape value) {
    if (_tickShape == value) {
      return;
    }
    _tickShape = value;
    markNeedsLayout();
  }

  SfTickShape get minorTickShape => _minorTickShape;
  SfTickShape _minorTickShape;

  set minorTickShape(SfTickShape value) {
    if (_minorTickShape == value) {
      return;
    }
    _minorTickShape = value;
    markNeedsLayout();
  }

  TextDirection get textDirection => _textDirection;
  TextDirection _textDirection;

  set textDirection(TextDirection value) {
    if (_textDirection == value) {
      return;
    }
    _textDirection = value;
    _updateTextPainter();
    markNeedsPaint();
  }

  MediaQueryData get mediaQueryData => _mediaQueryData;
  MediaQueryData _mediaQueryData;

  set mediaQueryData(MediaQueryData value) {
    if (_mediaQueryData == value) {
      return;
    }
    _mediaQueryData = value;
    _updateTextPainter();
    markNeedsLayout();
  }

  SfSliderThemeData get sliderThemeData => _sliderThemeData;
  SfSliderThemeData _sliderThemeData;

  set sliderThemeData(SfSliderThemeData value) {
    if (_sliderThemeData == value) {
      return;
    }
    _sliderThemeData = value;
    markNeedsPaint();
  }

  SfTooltipShape get tooltipShape => _tooltipShape;
  SfTooltipShape _tooltipShape;
  set tooltipShape(SfTooltipShape value) {
    if (_tooltipShape == value) {
      return;
    }
    _tooltipShape = value;
    markNeedsLayout();
  }

  void _updateTextPainter() {
    _textPainter
      ..textDirection = _textDirection
      ..textScaleFactor = _mediaQueryData.textScaleFactor;
  }

  @override
  void systemFontsDidChange() {
    super.systemFontsDidChange();
    _textPainter.markNeedsLayout();
    _updateTextPainter();
  }

  Offset _getTrackOffset() {
    final double dx = <double>[
          actualOverlaySize.width,
          actualThumbSize.width,
          actualTickSize.width,
          actualMinorTickSize.width
        ].reduce(math.max) /
        2;

    final double dy = <double>[
          actualOverlaySize.height,
          actualThumbSize.height,
          actualDivisorSize.height,
          _maxTrackHeight
        ].reduce(math.max) /
        2;

    return Offset(dx, dy);
  }

  double _getMaxTrackHeight() {
    return math.max(_sliderThemeData.activeTrackHeight,
        _sliderThemeData.inactiveTrackHeight);
  }

  String _getFormattedText(dynamic value) {
    if (isDateTime) {
      return _dateFormat != null ? _dateFormat.format(value) : value.toString();
    }
    return _numberFormat.format(value);
  }

  double _valueToFactor(dynamic value) {
    // If min and max are equal, the result will be NAN. This creates exception
    // and widget will not rendered.
    // So we have checked a condition (actualMax <= actualMin).
    final double factor = (value == null || actualMax <= actualMin)
        ? 0.0
        : ((value - actualMin) / (actualMax - actualMin));
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  double _valueToPosition(double value) {
    return _valueToFactor(value) * trackRect.width + trackRect.left;
  }

  @override
  void performLayout() {
    _actualHeight = math.max(
        2 * _trackOffset.dy,
        _trackOffset.dy +
            _maxTrackHeight / 2 +
            math.max(_actualTickHeight, _actualMinorTickHeight) +
            _actualLabelHeight);

    size = Size(
      constraints.hasBoundedWidth
          ? constraints.maxWidth
          : _minTrackWidth + 2 * _trackOffset.dx,
      constraints.hasBoundedHeight ? constraints.maxHeight : _actualHeight,
    );
    _generateLabelsAndMajorTicks();
    _generateMinorTicks();
  }

  @override
  void handleEvent(PointerEvent event, HitTestEntry entry) {
    // Checked [_isInteractionEnd] for avoiding multi touch.
    if (_isInteractionEnd && event.down && event is PointerDownEvent) {
      _dragGestureRecognizer.addPointer(event);
      _tapGestureRecognizer.addPointer(event);
    }
  }

  void _generateLabelsAndMajorTicks() {
    _visibleLabels.clear();
    _unformattedLabels?.clear();
    _majorTickPositions.clear();

    String label;
    int valueInms;
    double labelPosition;
    if (interval != null && interval > 0) {
      dynamic currentValue = _min;
      _divisions = (isDateTime
                  ? _getDateTimeDifference(_min, _max, _dateIntervalType)
                  : _max - _min)
              .toDouble() /
          _interval;
      for (int i = 0; i <= _divisions; i++) {
        label = _labelFormatterCallback(
            currentValue, _getFormattedText(currentValue));

        if (isDateTime) {
          valueInms = currentValue.millisecondsSinceEpoch;
        }
        _visibleLabels.add(label);
        _unformattedLabels
            ?.add(isDateTime ? valueInms.toDouble() : currentValue);

        labelPosition = _valueToFactor(isDateTime ? valueInms : currentValue) *
            (trackRect.width);
        if (!_majorTickPositions.contains(labelPosition)) {
          _majorTickPositions.add(labelPosition);
        }

        currentValue = isDateTime
            ? _getNextDate(currentValue, _dateIntervalType, _interval)
            : currentValue + _interval;
      }
    } else if (_showTicks || _showLabels) {
      _divisions = 1.0;
      label = _labelFormatterCallback(_min, _getFormattedText(_min));
      _visibleLabels.add(label);
      _unformattedLabels
          ?.add(isDateTime ? _min.millisecondsSinceEpoch.toDouble() : _min);
      label = _labelFormatterCallback(_max, _getFormattedText(_max));
      _visibleLabels.add(label);
      _unformattedLabels
          ?.add(isDateTime ? _max.millisecondsSinceEpoch.toDouble() : _max);

      labelPosition = _valueToFactor(actualMin) * (trackRect.width);
      _majorTickPositions.add(labelPosition);
      labelPosition = _valueToFactor(actualMax) * (trackRect.width);
      _majorTickPositions.add(labelPosition);
    }
  }

  void _generateMinorTicks() {
    if (interval != null && interval > 0) {
      _minorTickPositions.clear();
      final int majorTicksCount = _majorTickPositions.length;
      if (_minorTicksPerInterval > 0) {
        if (isDateTime) {
          double minorTickPosition;
          DateTime nextDate = _getNextDate(_min, _dateIntervalType, _interval);
          DateTime currentActualDate =
              _getNextDate(nextDate, _dateIntervalType, -_interval);
          for (int i = 1; i <= majorTicksCount; i++) {
            // Need to divide the region based on _minorTicksPerInterval.
            // So, added 1 with _minorTicksPerInterval.
            final double intervalDiff = _getDateTimeDifference(
                    currentActualDate, nextDate, _dateIntervalType) /
                (_minorTicksPerInterval + 1);
            // On dividing region equally between two dates, then it will be
            // equal to 1.0.
            if (intervalDiff == 1.0) {
              // To get nextDate of minorTick, need to pass intervals.
              // So, iteration value is used as interval and is started with 1.
              for (double j = 1; j <= _minorTicksPerInterval; j++) {
                final DateTime nextMinorDate =
                    _getNextDate(currentActualDate, _dateIntervalType, j);
                minorTickPosition = _valueToPosition(
                        nextMinorDate.millisecondsSinceEpoch.toDouble()) -
                    trackRect.left;
                _minorTickPositions.add(minorTickPosition);
              }
            } else {
              final double minorPositionDiff =
                  (nextDate.millisecondsSinceEpoch -
                          currentActualDate.millisecondsSinceEpoch) /
                      (_minorTicksPerInterval + 1);
              for (int j = 1; j <= _minorTicksPerInterval; j++) {
                minorTickPosition = _valueToPosition(
                        currentActualDate.millisecondsSinceEpoch +
                            (j * minorPositionDiff)) -
                    trackRect.left;
                _minorTickPositions.add(minorTickPosition);
              }
            }
            currentActualDate = nextDate;
            nextDate =
                _getNextDate(currentActualDate, _dateIntervalType, interval);
          }
        } else {
          for (int i = 0; i <= majorTicksCount - 1; i++) {
            final double minorPositionDiff = ((i + 1 < majorTicksCount
                        ? _majorTickPositions[i + 1]
                        : trackRect.width) -
                    _majorTickPositions[i]) /
                (_minorTicksPerInterval + 1);
            for (int j = 1; j <= _minorTicksPerInterval; j++) {
              _minorTickPositions
                  .add(_majorTickPositions[i] + j * minorPositionDiff);
            }
          }
        }
      }
    }
  }

  dynamic _getValueFromPosition(double position) {
    double valueFactor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    if (_textDirection == TextDirection.rtl) {
      valueFactor = 1.0 - valueFactor;
    }

    final dynamic actualValue = _factorToValue(valueFactor);
    return _getActualValue(valueInDouble: actualValue);
  }

  // The [value] argument holds either [double] or [DateTime] value and the
  // [valueInDouble] argument holds either [double] value for
  // numeric range slider and date time value in milliseconds for
  // date range slider.
  dynamic _getActualValue({dynamic value, double valueInDouble}) {
    if (isDiscrete) {
      if (!isDateTime) {
        final double maxMinDiff = _max - _min;
        value = _factorToValue(
            ((_valueToFactor(valueInDouble ?? value) * (maxMinDiff / stepSize))
                        .round() /
                    (maxMinDiff / stepSize))
                .clamp(0.0, 1.0));
      } else {
        DateTime currentDate = _min;
        DateTime nextDate;
        value = valueInDouble ?? value.millisecondsSinceEpoch.toDouble();
        final double clamppedValue = value.clamp(actualMin, actualMax);

        for (double i = actualMin; i < actualMax;) {
          nextDate = DateTime(
              currentDate.year + stepDuration.years,
              currentDate.month + stepDuration.months,
              currentDate.day + stepDuration.days,
              currentDate.hour + stepDuration.days,
              currentDate.minute + stepDuration.minutes,
              currentDate.second + stepDuration.seconds);

          final double currentDateInms =
              currentDate.millisecondsSinceEpoch.toDouble();
          final double nextDateInms =
              nextDate.millisecondsSinceEpoch.toDouble();

          if (clamppedValue >= currentDateInms &&
              clamppedValue <= nextDateInms) {
            final double dateDiff = (nextDateInms - currentDateInms).abs();
            final double halfDateDiff = dateDiff / 2;
            final bool shouldMoveNextInterval =
                clamppedValue >= (currentDateInms + halfDateDiff);
            value = shouldMoveNextInterval ? nextDate : currentDate;
            break;
          }
          i = nextDateInms;
          currentDate = nextDate;
        }
      }
    }

    return isDateTime
        ? (value ?? DateTime.fromMillisecondsSinceEpoch(valueInDouble.toInt()))
        : value ?? valueInDouble;
  }

  double _factorToValue(double factor) {
    return factor * (actualMax - actualMin) + actualMin;
  }

  double _positionToFactor(double position) {
    final double factor =
        ((position - trackRect.left) / trackRect.width).clamp(0.0, 1.0);
    return (_textDirection == TextDirection.rtl) ? 1.0 - factor : factor;
  }

  void _drawLabelsTicksAndDivisors(
      PaintingContext context,
      Rect trackRect,
      Offset offset,
      Offset thumbCenter,
      Offset startThumbCenter,
      Offset endThumbCenter,
      Animation<double> stateAnimation,
      dynamic value,
      SfRangeValues values) {
    int _dateTimePos = 0;
    final double dx = trackRect.left;
    final double dy = trackRect.top;
    final double halfTrackHeight = trackRect.height / 2;
    final bool isActive = startThumbCenter != null
        ? offset.dx >= startThumbCenter.dx && offset.dx <= endThumbCenter.dx
        : offset.dx <= thumbCenter.dx;
    final double divisorRadius = _divisorShape
            .getPreferredSize(_sliderThemeData, isActive: isActive)
            .width /
        2;

    final double tickRadius =
        _tickShape.getPreferredSize(_sliderThemeData).width / 2;

    double textvalue = isDateTime ? 0.0 : _min;
    int minorTickIndex = 0;
    final double maxRange = isDateTime ? _divisions : _max;

    if (_divisions != null && _divisions > 0) {
      while (textvalue <= maxRange) {
        final double tickPosition = _majorTickPositions[_dateTimePos];

        // Drawing ticks.
        if (showTicks) {
          Offset actualTickOffset =
              Offset(dx + tickPosition, dy + trackRect.height) +
                  (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          if (_majorTickPositions[_dateTimePos] == 0.0) {
            actualTickOffset =
                Offset(dx + tickPosition + tickRadius, dy + trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          } else if (_majorTickPositions[_dateTimePos] == trackRect.width) {
            actualTickOffset =
                Offset(dx + tickPosition - tickRadius, dy + trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
          }

          _tickShape.paint(context, actualTickOffset, thumbCenter,
              startThumbCenter, endThumbCenter,
              parentBox: this,
              themeData: _sliderThemeData,
              currentValue: value,
              currentValues: values,
              enableAnimation: stateAnimation,
              textDirection: _textDirection);
        }

        if (_interval != null && _interval > 0) {
          // Drawing minor ticks.
          if (_minorTicksPerInterval > 0) {
            for (int j = 0; j < _minorTicksPerInterval; j++) {
              final double currentMinorTickPosition =
                  _minorTickPositions[minorTickIndex];
              minorTickIndex++;
              if (currentMinorTickPosition < trackRect.width &&
                  currentMinorTickPosition > 0) {
                final Offset actualTickOffset = Offset(
                        dx + currentMinorTickPosition, dy + trackRect.height) +
                    (_sliderThemeData.tickOffset ?? const Offset(0, 0));
                _minorTickShape.paint(context, actualTickOffset, thumbCenter,
                    startThumbCenter, endThumbCenter,
                    parentBox: this,
                    themeData: _sliderThemeData,
                    currentValue: value,
                    currentValues: values,
                    enableAnimation: stateAnimation,
                    textDirection: _textDirection);
              }
            }
          }

          // Drawing divisors.
          if (showDivisors) {
            Offset divisorCenter =
                Offset(dx + tickPosition, dy + halfTrackHeight);
            if (_majorTickPositions[_dateTimePos] == 0.0) {
              divisorCenter = Offset(
                  dx + tickPosition + divisorRadius, dy + halfTrackHeight);
            } else if (_majorTickPositions[_dateTimePos] == trackRect.width) {
              divisorCenter = Offset(
                  dx + tickPosition - divisorRadius, dy + halfTrackHeight);
            }
            _divisorShape.paint(context, divisorCenter, thumbCenter,
                startThumbCenter, endThumbCenter,
                parentBox: this,
                themeData: _sliderThemeData,
                currentValue: value,
                currentValues: values,
                enableAnimation: stateAnimation,
                textDirection: _textDirection);
          }
        }

        // Drawing label.
        if (showLabels) {
          final dynamic labelText = _visibleLabels[_dateTimePos];
          final bool isRTL = _textDirection == TextDirection.rtl;
          double offsetX = dx + tickPosition;
          if (labelPlacement == LabelPlacement.betweenTicks) {
            offsetX += ((_dateTimePos + 1 <= _divisions
                        ? _majorTickPositions[_dateTimePos + 1]
                        : (isRTL ? trackRect.left : trackRect.width)) -
                    tickPosition) /
                2;
            if (isRTL
                ? offsetX <= trackRect.left
                : offsetX - dx >= trackRect.width) {
              break;
            }
          }
          final Offset actualLabelOffset = Offset(
                offsetX,
                dy + trackRect.height + _actualTickHeight,
              ) +
              (_sliderThemeData.labelOffset ?? const Offset(0, 0));

          _drawText(context, actualLabelOffset, thumbCenter, startThumbCenter,
              endThumbCenter, labelText,
              parentBox: this,
              themeData: _sliderThemeData,
              currentValue: value,
              currentValues: values,
              enableAnimation: stateAnimation,
              textPainter: _textPainter,
              textDirection: _textDirection);
        }

        // When interval is not set but [showLabels], [showTicks] enabled,
        // we need to show labels/ticks on min, max values. So we used
        // interval as _max - _min.
        final dynamic intervalDiff = isDateTime
            ? 1
            : _interval != null && _interval > 0 ? interval : _max - _min;
        textvalue += intervalDiff;
        _dateTimePos += 1;
      }
    }
  }

  /// Get the minimum and maximum value based on the
  /// intervalType to find the exact range.
  int _getDateTimeDifference(
      dynamic min, dynamic max, DateIntervalType intervalType) {
    final Duration diff = max.difference(min);
    switch (intervalType) {
      case DateIntervalType.months:
        return ((max.year - min.year) * DateTime.monthsPerYear) +
            max.month -
            min.month;
        break;
      case DateIntervalType.days:
        return diff.inDays;
        break;
      case DateIntervalType.hours:
        return diff.inHours;
        break;
      case DateIntervalType.minutes:
        return diff.inMinutes;
        break;
      case DateIntervalType.seconds:
        return diff.inSeconds;
        break;
      default:
        return max.year - min.year;
        break;
    }
  }

  /// Get the date time label based on the interval and intervalType.
  DateTime _getNextDate(
      DateTime currentDate, DateIntervalType intervalType, double interval) {
    switch (intervalType) {
      case DateIntervalType.months:
        // Make the label start date will always be 1 other than first label.
        return DateTime(
            currentDate.year, currentDate.month + interval.ceil(), 1);
        break;
      case DateIntervalType.days:
        currentDate = currentDate.add(Duration(days: interval.ceil()));
        return DateTime(currentDate.year, currentDate.month, currentDate.day);
        break;
      case DateIntervalType.hours:
        currentDate = currentDate.add(Duration(hours: interval.ceil()));
        return DateTime(currentDate.year, currentDate.month, currentDate.day,
            currentDate.hour);
        break;
      case DateIntervalType.minutes:
        return currentDate.add(Duration(minutes: interval.ceil()));
        break;
      case DateIntervalType.seconds:
        return currentDate.add(Duration(seconds: interval.ceil()));
        break;
      default:
        return DateTime(currentDate.year + interval.ceil(), 1, 1);
        break;
    }
  }

  void _drawText(PaintingContext context, Offset center, Offset thumbCenter,
      Offset startThumbCenter, Offset endThumbCenter, String text,
      {RenderProxyBox parentBox,
      SfSliderThemeData themeData,
      dynamic currentValue,
      SfRangeValues currentValues,
      Animation<double> enableAnimation,
      TextPainter textPainter,
      TextDirection textDirection}) {
    bool isInactive;
    switch (textDirection) {
      case TextDirection.ltr:
        // Added this condition to check whether consider single thumb or
        // two thumbs for finding inactive range.
        isInactive = startThumbCenter != null
            ? center.dx < startThumbCenter.dx || center.dx > endThumbCenter.dx
            : center.dx > thumbCenter.dx;
        break;
      case TextDirection.rtl:
        isInactive = startThumbCenter != null
            ? center.dx > startThumbCenter.dx || center.dx < endThumbCenter.dx
            : center.dx < thumbCenter.dx;
        break;
    }

    final TextSpan textSpan = TextSpan(
      text: text,
      style: isInactive
          ? themeData.inactiveLabelStyle
          : themeData.activeLabelStyle,
    );
    textPainter.text = textSpan;
    textPainter.layout();
    textPainter.paint(
        context.canvas, Offset(center.dx - textPainter.width / 2, center.dy));
  }

  double get _adjustmentUnit => (actualMax - actualMin) / 10;

  dynamic get _semanticActionUnit => isDateTime
      ? stepDuration != null ? _stepDuration : _adjustmentUnit
      : stepSize ?? _adjustmentUnit;

  dynamic _getNextSemanticValue(dynamic value, dynamic semanticActionUnit,
      {double actualValue}) {
    if (isDateTime) {
      if (stepDuration == null) {
        if (actualValue != null) {
          return DateTime.fromMillisecondsSinceEpoch(
              (actualValue + semanticActionUnit)
                  .clamp(actualMin, actualMax)
                  .toInt());
        }
      } else {
        final DateTime nextDate = DateTime(
            value.year + semanticActionUnit.years,
            value.month + semanticActionUnit.months,
            value.day + semanticActionUnit.days,
            value.hour + semanticActionUnit.days,
            value.minute + semanticActionUnit.minutes,
            value.second + semanticActionUnit.seconds);

        final double nextDateInms = nextDate.millisecondsSinceEpoch.toDouble();
        return nextDateInms < actualMax ? nextDate : _max;
      }
    } else {
      return (value + semanticActionUnit).clamp(_min, _max);
    }
  }

  dynamic _getPrevSemanticValue(dynamic value, dynamic semanticActionUnit,
      {double actualValue}) {
    if (isDateTime) {
      if (stepDuration == null) {
        if (actualValue != null) {
          return DateTime.fromMillisecondsSinceEpoch(
              (actualValue - semanticActionUnit)
                  .clamp(actualMin, actualMax)
                  .toInt());
        }
      } else {
        final DateTime prevDate = DateTime(
            value.year - semanticActionUnit.years,
            value.month - semanticActionUnit.months,
            value.day - semanticActionUnit.days,
            value.hour - semanticActionUnit.days,
            value.minute - semanticActionUnit.minutes,
            value.second - semanticActionUnit.seconds);

        final double prevDateInms = prevDate.millisecondsSinceEpoch.toDouble();
        return prevDateInms > actualMin ? prevDate : _min;
      }
    } else {
      return (value - semanticActionUnit).clamp(_min, _max);
    }
  }
}
